"""Command-line interface."""

import click


@click.command()
@click.version_option()
def main() -> None:
    """SSB Sirius Dash."""


if __name__ == "__main__":
    main(prog_name="ssb-sirius-dash")  # pragma: no cover



import dash_bootstrap_components as dbc
import pandas as pd
import plotly.graph_objects as go
from dash import Input
from dash import Output
from dash import State
from dash import callback
from dash import dcc
from dash import html
from dash.exceptions import PreventUpdate

from ..kostra_r_wrapper import hb_method
from .modal_functions import sidebar_button

states_options = [
    {
        "aar": ("var-aar", "value"),
        "termin": ("var-termin", "value"),
        "nace": ("var-nace", "value"),
        "nspekfelt": ("var-nspekfelt", "value"),
    }
]

ident_options = [
    {
        "orgb": ("var-bedrift", "value"),
        "orgf": ("var-foretak", "value"),
    }
]


class HBMethodModule:
    def __init__(
        self, database, hb_get_data_func, selected_state_keys, selected_ident, variable
    ):
        self.database = database
        self.hb_get_data = hb_get_data_func
        self.callbacks(selected_state_keys, selected_ident, variable)

    def make_hb_data(
        self, data_df: pd.DataFrame, p_c: int, p_u: float, p_a: float, ident, variable
    ) -> pd.DataFrame:
        hb_result = hb_method(
            data=data_df,
            p_c=p_c,
            p_u=p_u,
            p_a=p_a,
            id_field_name=ident,
            x_1_field_name=variable,
            x_2_field_name=f"{variable}_1",
        )

        return hb_result.sort_values(by=["maxX"])

    def make_hb_figure(self, data, variable):
        x = data["maxX"]
        y = data["ratio"]
        z = data["upperLimit"]
        k = data["lowerLimit"]

        fig = go.Figure()
        fig.add_trace(
            go.Scatter(
                x=x,
                y=y,
                mode="markers",
                hovertext=data["id"],
                name="Observasjon",
                marker={
                    "color": data["outlier"],
                    "colorscale": [[0, "#3498DB"], [1, "yellow"]],
                },
            )
        )
        fig.add_trace(go.Scatter(x=x, y=z, name="Øvre grense", marker_color="red"))
        fig.add_trace(go.Scatter(x=x, y=k, name="Nedre grense", marker_color="red"))
        fig.update_layout(
            height=800,
            title_text="HB-metoden",
            plot_bgcolor="#1F2833",
            paper_bgcolor="#1F2833",
            font_color="white",
        )
        fig.update_xaxes(title=variable, range=[0, max(x) * 1.05])
        fig.update_yaxes(title="Forholdstallet")

        return fig

    def layout(self):
        infoboks = html.Div(
            [
                dbc.Modal(
                    [
                        dbc.ModalHeader(dbc.ModalTitle("HB-metoden")),
                        dbc.ModalBody(
                            "Output: All units are returned, but the HB method is only performed on the data set where units with both x1 and x2 not missing and greater than zero are included. In this data set, units with x1 = x2 are included in he HB method only if they cover less than 50 per cent of the number of units in the stratum."
                        ),
                        dbc.ModalFooter(
                            html.Button(
                                "Close", id="close", className="ms-auto", n_clicks=0
                            )
                        ),
                    ],
                    id="modal",
                    is_open=False,
                ),
            ]
        )

        return html.Div(
            [
                dbc.Modal(
                    [
                        dbc.ModalHeader(dbc.ModalTitle("HB-metoden")),
                        dbc.ModalBody(
                            [
                                dbc.Row(infoboks),
                                dbc.Row(
                                    [
                                        dbc.Col(
                                            dbc.Button(
                                                "Kjør HB-modell",
                                                id="hb_button",
                                                style={"height": "100%"},
                                            )
                                        ),
                                        dbc.Col(
                                            dbc.Button(
                                                "Åpne infoboks", id="open", n_clicks=0
                                            ),
                                        ),
                                    ]
                                ),
                                dbc.Row(
                                    [
                                        self._build_input_field(
                                            "Skriv inn pC",
                                            "hb_pC",
                                            20,
                                            0,
                                            None,
                                            20,
                                            "Parameter that controls the length of the confidence interval.",
                                        ),
                                        self._build_input_field(
                                            "Skriv inn pU",
                                            "hb_pU",
                                            0.5,
                                            0,
                                            1,
                                            0.5,
                                            "Parameter that adjusts for different level of the variables.",
                                        ),
                                        self._build_input_field(
                                            "Skriv inn pA",
                                            "hb_pA",
                                            0.05,
                                            0,
                                            1,
                                            0.05,
                                            "Parameter that adjusts for small differences between the median and the 1st or 3rd quartile.",
                                        ),
                                    ],
                                    style={"margin": "20px"},
                                ),
                                dbc.Row(
                                    dcc.Loading(dcc.Graph(id="hb_figure")),
                                    style={"margin": "20px"},
                                ),
                            ]
                        ),
                    ],
                    id="hb-modal",
                    size="xl",
                    fullscreen="xxl-down",
                ),
                sidebar_button("🥼", "HB-Metoden", "sidebar-hb-button"),
            ]
        )

    def _build_input_field(
        self,
        label,
        id_name,
        default_value,
        min_value,
        max_value,
        initial_value,
        tooltip_text,
    ):
        return dbc.Col(
            dbc.Stack(
                [
                    html.Span(
                        label,
                        id=f"{id_name}_tooltip",
                        style={
                            "textDecoration": "underline",
                            "cursor": "pointer",
                            "color": "white",
                        },
                    ),
                    dbc.Tooltip(
                        [
                            html.P(tooltip_text),
                        ],
                        target=f"{id_name}_tooltip",
                    ),
                    dcc.Input(
                        id=id_name,
                        type="number",
                        value=default_value,
                        min=min_value,
                        max=max_value,
                    ),
                ]
            )
        )

    def callbacks(self, selected_state_keys, selected_ident, variable):
        states_dict = states_options[0]
        dynamic_states = [
            State(states_dict[key][0], states_dict[key][1])
            for key in selected_state_keys
        ]

        ident = ident_options[0][selected_ident]
        component_id, property_name = ident
        output_object = Output(component_id, property_name, allow_duplicate=True)

        @callback(
            Output("hb_figure", "figure"),
            Input("hb_button", "n_clicks"),
            State("hb_pC", "value"),
            State("hb_pU", "value"),
            State("hb_pA", "value"),
            *dynamic_states,
        )
        def use_hb(n_click, pC, pU, pA, *dynamic_states):
            states_values = dynamic_states[: len(selected_state_keys)]
            state_params = {
                key: value for key, value in zip(selected_state_keys, states_values)
            }

            args = []
            for key in selected_state_keys:
                var = state_params.get(key)
                if var is not None:
                    args.append(var)

            if "nace" in state_params and state_params["nace"] is not None:
                n = len(state_params["nace"])
                args.append(n)

            if n_click:
                data = self.hb_get_data(self.database, *args)
                data = self.make_hb_data(data, pC, pU, pA, selected_ident, variable)
                return self.make_hb_figure(data, variable)
            else:
                raise PreventUpdate

        @callback(
            Output("hb-modal", "is_open"),
            Input("sidebar-hb-button", "n_clicks"),
            State("hb-modal", "is_open"),
        )
        def sqlmodal_toggle(n, is_open):
            if n:
                return not is_open
            return is_open

        @callback(
            output_object,
            Input("hb_figure", "clickData"),
            prevent_initial_call=True,
        )
        def hb_to_main_table(clickdata):
            if clickdata:
                ident = clickdata["points"][0]["hovertext"]
                return ident



import logging
import timeit
from typing import Optional

import pandas as pd
from rpy2.robjects import conversion
from rpy2.robjects import default_converter
from rpy2.robjects import pandas2ri
from rpy2.robjects.packages import InstalledSTPackage
from rpy2.robjects.packages import importr

logger = logging.getLogger(__name__)

# Global variable to store the R package Kostra
_kostra_r: Optional[InstalledSTPackage] = None


def _get_kostra_r() -> InstalledSTPackage:
    """Loads the R package Kostra.

    :return: Kostra R package
    """
    if _kostra_r is not None:
        return _kostra_r

    start_time = timeit.default_timer()
    globals()["_kostra_r"] = importr("Kostra")
    logger.info(
        "Finished loading Kostra in %3g seconds", (timeit.default_timer() - start_time)
    )
    return globals()["_kostra_r"]


def th_error(
    data: pd.DataFrame, id_field_name: str, x_1_field_name: str, x_2_field_name: str
) -> pd.DataFrame:
    """Runs the ThError method from the R package Kostra.

    :param data: The data to run the method on
    :param id_field_name: The name of the id field
    :param x_1_field_name: The name of the first x field
    :param x_2_field_name: The name of the second x field
    :return: The result of the method
    """
    with conversion.localconverter(default_converter + pandas2ri.converter):
        th_error_result = _get_kostra_r().ThError(
            data=data, id=id_field_name, x1=x_1_field_name, x2=x_2_field_name
        )
        return th_error_result[th_error_result.outlier == 0]


def hb_method(
    data: pd.DataFrame,
    p_c: int,
    p_u: float,
    p_a: float,
    id_field_name: str = "id",
    x_1_field_name: str = "x1",
    x_2_field_name: str = "x2",
) -> pd.DataFrame:
    """Runs the Hb method from the R package Kostra.

    :param data: The data to run the method on
    :param p_c: The value of pC
    :param p_u: The value of pU
    :param p_a: The value of pA
    :param id_field_name: The name of the id field
    :param x_1_field_name: The name of the first x field
    :param x_2_field_name: The name of the second x field
    :return: The result of the method
    """
    with conversion.localconverter(default_converter + pandas2ri.converter):
        return _get_kostra_r().Hb(
            data=data,
            id=id_field_name,
            x1=x_1_field_name,
            x2=x_2_field_name,
            pC=p_c,
            pU=p_u,
            pA=p_a,
        )



import dash_bootstrap_components as dbc
from dash import html


def main_layout(modal_list, tab_list, variable_list):
    selected_tab_list = [dbc.Tab(tab.layout(), label=tab.label) for tab in tab_list]
    layout = dbc.Container(
        [
            html.Div(
                id="notifications-container",
                style={"position": "fixed", "z-index": 9999},
            ),
            html.P(
                id="update-status", style={"font-size": "60%", "visibility": "hidden"}
            ),
            html.Div(
                id="main-layout",
                style={
                    "height": "100vh",
                    "overflow": "hidden",
                    "display": "grid",
                    "grid-template-columns": "5% 95%",
                },
                children=[
                    html.Div(
                        className="bg-secondary",
                        style={
                            "display": "flex",
                            "flex-direction": "column",
                            "height": "100%",
                        },
                        children=modal_list,
                    ),
                    html.Div(
                        children=[
                            dbc.Row(children=variable_list),
                            dbc.Tabs(
                                selected_tab_list,
                            ),
                        ],
                    ),
                ],
            ),
        ],
        fluid=True,
        className="dbc dbc-ag-grid",
    )
    return layout



import dash_bootstrap_components as dbc
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from dash import Input
from dash import Output
from dash import State
from dash import callback
from dash import dcc
from dash import html
from dash.exceptions import PreventUpdate

from .modal_functions import sidebar_button


class KvalitetsindikatorerModule:
    """indicators: list
    List of indicators to include.
    """

    def __init__(self, indicators):

        self.indicators = indicators
        self.callbacks()

    def layout(self):
        return html.Div(
            [
                dbc.Modal(
                    [
                        dbc.ModalHeader([dbc.ModalTitle("kvalitetsindikatorer")]),
                        dbc.ModalBody(
                            [
                                html.Div(
                                    [indicator.card for indicator in self.indicators],
                                    style={
                                        "display": "grid",
                                        "gridTemplateColumns": "repeat(auto-fit, minmax(200px, 1fr))",
                                        "gridGap": "20px",
                                    },
                                )
                            ]
                        ),
                    ],
                    id="kvalitetsindikatorer-modal",
                    size="xl",
                    fullscreen="xxl-down",
                ),
                sidebar_button(
                    "🎯", "Kvalitetsindikatorer", "sidebar-kvalitetsindikatorer-button"
                ),
            ]
        )

    def callbacks(self):
        @callback(
            Output("kvalitetsindikatorer-modal", "is_open"),
            Input("sidebar-kvalitetsindikatorer-button", "n_clicks"),
            State("kvalitetsindikatorer-modal", "is_open"),
        )
        def makrotabellmodal_toggle(n, is_open):
            if n:
                return not is_open
            return is_open


class KvalitetsindikatorEditeringsandel:
    """ """

    def __init__(
        self,
        get_current_data_func,
        get_change_data_func,
        periode,
        var_name,
        grouping_vars,
        ident_var,
        database,
        key_vars=None,
    ):
        self.periode = periode
        self.ident_var = ident_var
        self.var_name = var_name
        self.grouping_vars = grouping_vars
        self.get_current_data = get_current_data_func
        self.get_change_data = get_change_data_func
        self.database = database
        if key_vars:
            self.key_vars = key_vars  # TODO

        self.callbacks()

        self.card = html.Div(
            [
                dbc.Card(
                    [
                        dbc.CardBody(
                            [
                                html.H5("1 - Editeringsandel", className="card-title"),
                                dcc.Graph(
                                    figure=go.Figure(
                                        go.Indicator(
                                            mode="number+delta",
                                            value=self.editeringsandel(
                                                self.database, self.periode
                                            ),
                                            number={"prefix": ""},
                                            # delta={"position": "bottom", "reference": self.editeringsandel(self.periode-1)}, # TODO
                                            domain={"x": [0, 1], "y": [0, 1]},
                                        )
                                    ).update_layout(
                                        height=150,
                                        margin=dict(l=20, r=20, t=20, b=20),
                                    ),
                                    config={"displayModeBar": False},
                                ),
                            ]
                        ),
                        dbc.CardFooter(
                            dbc.Button(
                                "Detaljer",
                                id="kvalitet-editeringsandel-button-details",
                            )
                        ),
                    ],
                    style={
                        "width": "18rem",
                        "margin": "10px",
                    },
                ),
                dbc.Modal(
                    [
                        dbc.ModalHeader("1 - Editeringsandel"),
                        dbc.ModalBody(
                            [
                                html.Div(
                                    children=[
                                        dcc.Dropdown(
                                            id="kvalitet-editeringsandel-dropdown",
                                            options=[
                                                {"label": x, "value": x}
                                                for x in [var_name, *grouping_vars]
                                            ],
                                        )
                                    ],
                                ),
                                dcc.Loading(
                                    id="kvalitet-editeringsandel-details",
                                ),
                            ]
                        ),
                    ],
                    id="editeringsandel-modal",
                ),
            ]
        )

    def editeringsandel(self, database, periode):
        total = pd.DataFrame(
            self.get_current_data(database, periode).agg({self.ident_var: "nunique"})
        )
        changes = pd.DataFrame(
            self.get_change_data(database, periode).agg({self.ident_var: "nunique"})
        )
        return (changes / total * 100).iloc[0][0]

    def editeringsandel_details(self, group, database, periode):
        if isinstance(group, str):
            group = [group]
        total = (
            self.get_current_data(database, periode, group)
            .groupby(group)
            .agg({self.ident_var: "nunique"})
            .rename(columns={self.ident_var: "units"})
        )
        changes = (
            self.get_change_data(database, periode, group)
            .groupby(group)
            .agg({self.ident_var: "nunique"})
            .rename(columns={self.ident_var: "edited_units"})
        )
        c = pd.merge(total, changes, on=group, how="left").fillna(0)
        c["editeringsandel"] = c["edited_units"] / c["units"] * 100
        return c.reset_index()

    def callbacks(self):
        @callback(
            Output("editeringsandel-modal", "is_open"),
            Input("kvalitet-editeringsandel-button-details", "n_clicks"),
            State("editeringsandel-modal", "is_open"),
        )
        def kvalitetediteringsandel_modaltoggle(n, is_open):
            if n:
                return not is_open
            return is_open

        @callback(
            Output("kvalitet-editeringsandel-details", "children"),
            Input("kvalitet-editeringsandel-dropdown", "value"),
        )
        def editeringsandel_detailed(grouping_var):
            if grouping_var:
                detail_data = self.editeringsandel_details(
                    grouping_var, self.database, self.periode
                )
                return dcc.Graph(
                    figure=px.bar(
                        detail_data,
                        x="editeringsandel",
                        y=grouping_var,
                        orientation="h",
                    )
                )
            else:
                raise PreventUpdate


class KvalitetsindikatorKontrollutslagsandel:
    def __init__(
        self,
        kontrolldokumentasjon,
    ):
        """Kontrolldokumentasjon skal være et datasett med kolonnene:
        kontroll_id | Enheter kontrollert | Kontrollutslag
        """
        self.kontrolldokumentasjon = kontrolldokumentasjon
        self.kontrollutslagsandel_total, self.kontrollutslagsandel_detaljer = (
            self.kontrollutslag()
        )

        self.callbacks()

        self.card = html.Div(
            [
                dbc.Card(
                    [
                        dbc.CardBody(
                            [
                                html.H5(
                                    "2 - Kontrollutslagsandel", className="card-title"
                                ),
                                dcc.Graph(
                                    figure=go.Figure(
                                        go.Indicator(
                                            mode="number+delta",
                                            value=self.kontrollutslagsandel_total,
                                            number={"prefix": ""},
                                            # delta={"position": "bottom", "reference": self.editeringsandel(self.periode-1)}, # TODO
                                            domain={"x": [0, 1], "y": [0, 1]},
                                        )
                                    ).update_layout(
                                        height=150,
                                        margin=dict(l=20, r=20, t=20, b=20),
                                    ),
                                    config={"displayModeBar": False},
                                ),
                            ]
                        ),
                        dbc.CardFooter(
                            dbc.Button(
                                "Detaljer",
                                id="kvalitet-kontrollutslagsandel-button-details",
                            )
                        ),
                    ],
                    style={
                        "width": "18rem",
                        "margin": "10px",
                    },
                ),
                dbc.Modal(
                    [
                        dbc.ModalHeader("2 - Kontrollutslagsandel"),
                        dbc.ModalBody(
                            [
                                html.Div(
                                    children=[
                                        dag.AgGrid(
                                            columnDefs=[
                                                {"field": x}
                                                for x in self.kontrollutslagsandel_detaljer.columns
                                            ],
                                            rowData=self.kontrollutslagsandel_detaljer.to_dict(
                                                "records"
                                            ),
                                        )
                                    ],
                                ),
                                dcc.Loading(
                                    id="kvalitet-kontrollutslagsandel-details",
                                ),
                            ]
                        ),
                    ],
                    id="kontrollutslagsandel-modal",
                ),
            ]
        )

    def kontrollutslag(self):
        total = (
            self.kontrolldokumentasjon["Kontrollutslag"].sum()
            / self.kontrolldokumentasjon["Enheter kontrollert"].sum()
        )

        self.kontrolldokumentasjon["kontrollutslagsandel"] = (
            self.kontrolldokumentasjon["Kontrollutslag"]
            / self.kontrolldokumentasjon["Enheter kontrollert"]
        )

        return total, self.kontrolldokumentasjon

    def callbacks(self):
        @callback(
            Output("kontrollutslagsandel-modal", "is_open"),
            Input("kvalitet-kontrollutslagsandel-button-details", "n_clicks"),
            State("kontrollutslagsandel-modal", "is_open"),
        )
        def kvalitetkontrollutslagsandel_modaltoggle(n, is_open):
            if n:
                return not is_open
            return is_open


class KvalitetsindikatorEffektaveditering:
    """ """

    def __init__(
        self,
        get_current_data_func,
        get_original_data_func,
        periode,
        key_vars,
        grouping_vars,
        database,
    ):
        self.get_current_data = get_current_data_func
        self.get_original_data = get_original_data_func
        self.periode = periode
        self.key_vars = key_vars
        self.grouping_vars = grouping_vars
        self.database = database

        self.callbacks()

        self.card = html.Div(
            [
                dbc.Card(
                    [
                        dbc.CardBody(
                            [
                                html.H5(
                                    "4 - Effekten av editering", className="card-title"
                                ),
                                dcc.Graph(
                                    figure=go.Figure(
                                        go.Indicator(
                                            mode="number+delta",
                                            value=self.get_comparison_data(
                                                self.periode
                                            )["effekt av editering"][0],
                                            number={"prefix": ""},
                                            # delta={"position": "bottom", "reference": self.editeringsandel(self.periode-1)}, # TODO
                                            domain={"x": [0, 1], "y": [0, 1]},
                                        )
                                    ).update_layout(
                                        height=150,
                                        margin=dict(l=20, r=20, t=20, b=20),
                                    ),
                                    config={"displayModeBar": False},
                                ),
                            ]
                        ),
                        dbc.CardFooter(
                            dbc.Button(
                                "Detaljer",
                                id="kvalitet-effekt-button-details",
                            )
                        ),
                    ],
                    style={
                        "width": "18rem",
                        "margin": "10px",
                    },
                ),
                dbc.Modal(
                    [
                        dbc.ModalHeader("4 - Effekten av editering"),
                        dbc.ModalBody(
                            [
                                html.Div(
                                    children=[
                                        dcc.Dropdown(
                                            id="kvalitet-effekt-dropdown",
                                            options=[
                                                {"label": x, "value": x}
                                                for x in [*grouping_vars]
                                            ],
                                        )
                                    ],
                                ),
                                dcc.Loading(
                                    id="kvalitet-effekt-details",
                                ),
                            ]
                        ),
                    ],
                    id="effekt-modal",
                ),
            ]
        )

    def get_comparison_data(self, periode, grouping=None):
        if grouping is None:
            grouping = []
        elif isinstance(grouping, str):
            grouping = [grouping]
        edited = (
            self.get_current_data(self.database, periode)
            .melt(id_vars=["oppgavegivernummer", *grouping], value_vars=self.key_vars)
            .groupby([*grouping, "variable"])
            .agg({"value": "sum"})
            .rename(columns={"value": "editert"})
            .reset_index()
        )

        ueditert = (
            self.get_original_data(self.database, periode)
            .melt(id_vars=["oppgavegivernummer", *grouping], value_vars=self.key_vars)
            .groupby([*grouping, "variable"])
            .agg({"value": "sum"})
            .rename(columns={"value": "ueditert"})
            .reset_index()
        )

        merged = edited.merge(ueditert, on=[*grouping, "variable"])

        merged["effekt av editering"] = (
            (merged["ueditert"] - merged["editert"]) / merged["editert"] * 100
        )

        return merged

    def callbacks(self):
        @callback(
            Output("effekt-modal", "is_open"),
            Input("kvalitet-effekt-button-details", "n_clicks"),
            State("effekt-modal", "is_open"),
        )
        def kvaliteteffekt_modaltoggle(n, is_open):
            if n:
                return not is_open
            return is_open

        @callback(
            Output("kvalitet-effekt-details", "children"),
            Input("kvalitet-effekt-dropdown", "value"),
        )
        def kvalitet_effekt_detailed(grouping_var):
            if grouping_var:
                detail_data = self.get_comparison_data(self.periode, grouping_var)
                return dcc.Graph(
                    figure=px.bar(
                        detail_data,
                        x="effekt av editering",
                        y=grouping_var,
                        orientation="h",
                    )
                )
            else:
                raise PreventUpdate



import dash_ag_grid as dag
import dash_bootstrap_components as dbc
import pandas as pd
from dash import Input
from dash import Output
from dash import State
from dash import callback
from dash import html

from ssb_sirius_dash import sidebar_button

from ..control.framework import lag_kontroll_dokumentasjon

# +
ident_options = [
    {
        "orgb": ("var-bedrift", "value"),
        "orgf": ("var-foretak", "value"),
    }
]


class Kontroller:
    def __init__(self, ident, kontroll_dokumentasjon_path=None):
        self.ident = ident_options[0][ident]
        if kontroll_dokumentasjon_path:
            import json

            import dapla as dp

            with dp.FileClient.gcs_open(kontroll_dokumentasjon_path, "r") as outfile:
                data = json.load(outfile)
            self.kontrolltabell = lag_kontroll_dokumentasjon(data)
            self.utslagstabell = pd.DataFrame(data["kontrollutslag"])
        self.callbacks()

    def layout(self):
        return html.Div(
            [
                dbc.Modal(
                    [
                        dbc.ModalHeader([dbc.ModalTitle("⚠️ Kontroller")]),
                        dbc.ModalBody(
                            [
                                dag.AgGrid(
                                    id="kontroll-table-overview",
                                    columnDefs=[
                                        {"field": x}
                                        for x in [
                                            "kontroll_id",
                                            "kontrolltype",
                                            "feilbeskrivelse",
                                            "Enheter i datasettet",
                                            "Enheter kontrollert",
                                            "Kontrollutslag",
                                        ]
                                    ],
                                    rowData=self.kontrolltabell.to_dict(
                                        orient="records"
                                    ),
                                ),
                                dag.AgGrid(
                                    id="kontroll-table-detailed",
                                    columnDefs=[
                                        {"field": x}
                                        for x in ["observasjon_id", "feilbeskrivelse"]
                                    ]
                                    + [
                                        {
                                            "field": "kontrollnavn",
                                            "hide": True,
                                            "filter": True,
                                        }
                                    ],
                                    rowData=self.utslagstabell.to_dict(
                                        orient="records"
                                    ),
                                ),
                            ]
                        ),
                    ],
                    id="kontroller-modal",
                    size="xl",
                    fullscreen="xxl-down",
                ),
                sidebar_button("⚠️", "Kontroller", "sidebar-kontroller-button"),
            ]
        )

    def callbacks(self):
        @callback(
            Output("kontroller-modal", "is_open"),
            Input("sidebar-kontroller-button", "n_clicks"),
            State("kontroller-modal", "is_open"),
        )
        def kontrollermodal_toggle(n, is_open):
            if n:
                return not is_open
            return is_open

        @callback(
            Output("kontroll-table-detailed", "filterModel"),
            Input("kontroll-table-overview", "cellClicked"),
            State("kontroll-table-overview", "rowData"),
        )
        def kontroll_main_click(click, rowData):
            print(click)
            kontroll = rowData[click["rowIndex"]]["kontroll_id"]
            return {
                "kontrollnavn": {
                    "filterType": "text",
                    "type": "contains",
                    "filter": kontroll,
                }
            }

        @callback(
            Output(self.ident[0], self.ident[1]),
            Input("kontroll-table-detailed", "cellClicked"),
            State("kontroll-table-detailed", "rowData"),
        )
        def kontroll_detail_click(click, rowData):
            return rowData[click["rowIndex"]]["observasjon_id"]



import dash_ag_grid as dag
import dash_bootstrap_components as dbc
import pandas as pd
import plotly.express as px
from dash import callback
from dash import dcc
from dash import html
from dash.dependencies import Input
from dash.dependencies import Output
from dash.dependencies import State

from .modal_functions import sidebar_button


class VisualiseringsbyggerModule:
    def __init__(self, database):
        self.database = database
        self.callbacks()

    def layout(self):
        layout = html.Div(
            [
                dbc.Modal(
                    [
                        dbc.ModalHeader(
                            dbc.ModalTitle("🏗️ Visualiseringsbygger"),
                        ),
                        dbc.ModalBody(
                            [
                                html.Div(
                                    style={
                                        "display": "grid",
                                        "height": "100%",
                                        "grid-template-rows": "5% 25% 35% 5% 50%",
                                    },
                                    children=[
                                        html.Div(
                                            dbc.Button(
                                                "Kjør spørring",
                                                id="sql-button",
                                                style={
                                                    "display": "flex",
                                                    "flex-direction": "column",
                                                    "align-items": "center",
                                                    "word-break": "break-all",
                                                    "margin-bottom": "5%",
                                                    "width": "100%",
                                                },
                                            )
                                        ),
                                        html.Div(
                                            dcc.Textarea(
                                                id="sqlmodal-textarea",
                                                placeholder="SELECT * FROM databasetabell",
                                                style={"width": "100%", "height": 300},
                                            ),
                                        ),
                                        html.Div(
                                            dag.AgGrid(
                                                id="sql-output-table",
                                                className="ag-theme-alpine-dark header-style-on-filter",
                                                columnSize="responsiveSizeToFit",
                                                style={"pagination": True},
                                            )
                                        ),
                                        html.Div(
                                            style={
                                                "display": "grid",
                                                "height": "100%",
                                                "grid-template-columns": "25% 25% 25% 25%",
                                            },
                                            children=[
                                                dcc.Dropdown(
                                                    id="sql-x",
                                                    placeholder="x-aksen",
                                                    value=[],
                                                    multi=True,
                                                    className="dbc",
                                                ),
                                                dcc.Dropdown(
                                                    id="sql-y",
                                                    placeholder="y-aksen",
                                                    value=[],
                                                    multi=True,
                                                    className="dbc",
                                                ),
                                                dcc.Dropdown(
                                                    id="sql-hover",
                                                    placeholder="hoverdata",
                                                    value=[],
                                                    multi=True,
                                                    className="dbc",
                                                ),
                                                dcc.Dropdown(
                                                    id="sql-graph-type",
                                                    placeholder="graftype",
                                                    options=[
                                                        {
                                                            "label": "scatter",
                                                            "value": "scatter",
                                                        },
                                                        {
                                                            "label": "box",
                                                            "value": "box",
                                                        },
                                                        {
                                                            "label": "violin",
                                                            "value": "violin",
                                                        },
                                                        {
                                                            "label": "line",
                                                            "value": "line",
                                                        },
                                                        {
                                                            "label": "bar",
                                                            "value": "bar",
                                                        },
                                                        {
                                                            "label": "histogram",
                                                            "value": "histogram",
                                                        },
                                                    ],
                                                    className="dbc",
                                                ),
                                            ],
                                        ),
                                        dcc.Loading(
                                            id="sql-graph1-loading",
                                            children=[
                                                dcc.Graph(
                                                    id="sql-graph1", className="m-4"
                                                )
                                            ],
                                            type="graph",
                                            style={
                                                "position": "fixed",
                                                "z-index": 9999,
                                            },
                                        ),
                                    ],
                                )
                            ],
                        ),
                    ],
                    id="sql-modal",
                    size="xl",
                    fullscreen="xxl-down",
                ),
                sidebar_button("🏗️", "Visualiseringsbygger", "sidebar-sql-button"),
            ],
        )
        return layout

    def callbacks(self):
        @callback(
            Output("sql-modal", "is_open"),
            Input("sidebar-sql-button", "n_clicks"),
            State("sql-modal", "is_open"),
        )
        def sqlmodal_toggle(n, is_open):
            if n:
                return not is_open
            return is_open

        @callback(
            Output("sql-output-table", "rowData"),
            Output("sql-output-table", "columnDefs"),
            Output("sql-x", "options"),
            Output("sql-y", "options"),
            Output("sql-hover", "options"),
            Input("sql-button", "n_clicks"),
            State("sqlmodal-textarea", "value"),
        )
        def sql_query(n_clicks, value):
            if n_clicks > 0:
                df = self.database.query(f"""{value}""")
                options = [{"label": col, "value": col} for col in df.columns]
                columns = [{"headerName": col, "field": col} for col in df.columns]
                return df.to_dict("records"), columns, options, options, options

        @callback(
            Output("sql-graph1", "figure"),
            Input("sql-x", "value"),
            Input("sql-y", "value"),
            Input("sql-hover", "value"),
            Input("sql-graph-type", "value"),
            State("sql-output-table", "rowData"),
            State("sql-output-table", "columnDefs"),
        )
        def update_graph(x_axis, y_axis, hover_data, graph_type, rowData, columnDefs):
            if x_axis and y_axis:
                if isinstance(x_axis, list) and len(x_axis) == 1:
                    x_axis = x_axis[0]
                if isinstance(y_axis, list) and len(y_axis) == 1:
                    y_axis = y_axis[0]
                if (
                    isinstance(x_axis, list)
                    and len(x_axis) > 1
                    and isinstance(y_axis, list)
                    and len(y_axis) > 1
                ):
                    y_axis = y_axis[0]
                df = pd.DataFrame(rowData)
                columns = [col["field"] for col in columnDefs]
                df.columns = columns
                df.fillna(0)
                if graph_type == "scatter":
                    fig = px.scatter(
                        df,
                        x=x_axis,
                        y=y_axis,
                        hover_data=[hover_data] if hover_data else None,
                    )
                elif graph_type == "line":
                    fig = px.line(
                        df,
                        x=x_axis,
                        y=y_axis,
                        hover_data=[hover_data] if hover_data else None,
                    )
                elif graph_type == "bar":
                    fig = px.bar(
                        df,
                        x=x_axis,
                        y=y_axis,
                        hover_data=[hover_data] if hover_data else None,
                    )
                elif graph_type == "box":
                    fig = px.box(df, x=x_axis, y=y_axis, points="all")
                elif graph_type == "violin":
                    fig = px.violin(df, x=x_axis, y=y_axis, box=True, points="all")
                elif graph_type == "histogram":
                    fig = px.histogram(
                        df,
                        x=x_axis,
                        y=y_axis,
                        hover_data=[hover_data] if hover_data else None,
                    )
                else:
                    fig = {}
                return fig
            else:
                return {}



import ast

import dash_ag_grid as dag
import dash_bootstrap_components as dbc
from dash import callback
from dash import html
from dash.dependencies import Input
from dash.dependencies import Output
from dash.dependencies import State


class FrisøkTab:
    def __init__(self, database):
        self.database = database
        self.label = "🔍 Frisøk"

    def layout(self):
        layout = html.Div(
            [
                html.Div(
                    dbc.Textarea(
                        id="tab-frisøk-textarea1",
                        size="xxl",
                        placeholder="SELECT * FROM databasetabell",
                    ),
                ),
                html.Div(
                    style={"display": "grid", "grid-template-columns": "80% 20%"},
                    children=[
                        dbc.Input(
                            id="tab-frisøk-input1",
                            placeholder="Velg partisjoner. f.eks. {'aar': [2023], 'termin':[1, 2]}",
                        ),
                        dbc.Button(
                            "kjør",
                            id="tab-frisøk-button1",
                        ),
                    ],
                ),
                dag.AgGrid(
                    defaultColDef={"editable": True},
                    id="tab-frisøk-table1",
                    className="ag-theme-alpine-dark header-style-on-filter",
                ),
            ]
        )
        return layout

    def callbacks(self):
        @callback(
            Output("tab-frisøk-table1", "rowData"),
            Output("tab-frisøk-table1", "columnDefs"),
            Input("tab-frisøk-button1", "n_clicks"),
            State("tab-frisøk-textarea1", "value"),
            State("tab-frisøk-input1", "value"),
        )
        def table_frisøk(n_clicks, query, partisjoner):
            if n_clicks > 0:
                if partisjoner is not None:
                    partisjoner = ast.literal_eval(partisjoner)
                df = self.database.query(query, partition_select=partisjoner)
                columns = [
                    {
                        "headerName": col,
                        "field": col,
                        "hide": True if col == "row_id" else False,
                    }
                    for col in df.columns
                ]
                return df.to_dict("records"), columns



import uuid
from typing import Any
from typing import Optional

import dash_bootstrap_components as dbc
import pandas as pd
import plotly.express as px
from dash import MATCH
from dash import Input
from dash import Output
from dash import State
from dash import callback
from dash import dcc
from dash import html
from plotly.graph_objs import Figure
from plotly.graph_objs import Scatter
from plotly.graph_objs import scatter

from .kostra_r_wrapper import hb_method
from .kostra_r_wrapper import th_error


def _generate_hb_aio_id(subcomponent: str, aio_id: Any) -> dict[str, Any]:
    return {"component": "HbMethodAIO", "subcomponent": subcomponent, "aio_id": aio_id}


#
# component ids
#
HB_DATA_STORE_ID = "data_store"
HB_SCATTERPLOT_ID = "scatterplot"
HB_PC_ID = "p_c"
HB_PU_ID = "p_u"
HB_PA_ID = "p_a"
HB_FILTER_OP_ID = "hb_filter_op"
HB_FILTER_VALUE_ID = "hb_filter_value"


class HbMethodAIO(html.Div):  # type: ignore
    """HbMethodAIO is an All-in-One component that is composed of a parent `html.Div`."""

    class Ids:
        """A set of functions that create pattern-matching IDs of the subcomponents."""

        @staticmethod
        def data_store(aio_id: Any) -> dict[str, Any]:
            """Returns the component ID for the data store.

            :param aio_id: Component aio-ID
            :return: Component ID
            """
            return _generate_hb_aio_id(HB_DATA_STORE_ID, aio_id)

        @staticmethod
        def scatterplot(aio_id: Any) -> dict[str, Any]:
            """Returns the component ID for the scatterplot.

            :param aio_id: Component aio-ID
            :return: Component ID
            """
            return _generate_hb_aio_id(HB_SCATTERPLOT_ID, aio_id)

        @staticmethod
        def p_c(aio_id: Any) -> dict[str, Any]:
            """Returns the component ID for the pC dropdown.

            :param aio_id: Component aio-ID
            :return: Component ID
            """
            return _generate_hb_aio_id(HB_PC_ID, aio_id)

        @staticmethod
        def p_u(aio_id: Any) -> dict[str, Any]:
            """Returns the component ID for the pU dropdown.

            :param aio_id: Component aio-ID
            :return: Component ID
            """
            return _generate_hb_aio_id(HB_PU_ID, aio_id)

        @staticmethod
        def p_a(aio_id: Any) -> dict[str, Any]:
            """Returns the component ID for the pA dropdown.

            :param aio_id: Component aio-ID
            :return: Component ID
            """
            return _generate_hb_aio_id(HB_PA_ID, aio_id)

        @staticmethod
        def hb_filter_op(aio_id: Any) -> dict[str, Any]:
            """Returns the component ID for the filter operation dropdown.

            :param aio_id: Component aio-ID
            :return: Component ID
            """
            return _generate_hb_aio_id(HB_FILTER_OP_ID, aio_id)

        @staticmethod
        def hb_filter_value(aio_id: Any) -> dict[str, Any]:
            """Returns the component ID for the filter value dropdown.

            :param aio_id: Component aio-ID
            :return: Component ID
            """
            return _generate_hb_aio_id(HB_FILTER_VALUE_ID, aio_id)

    ids = Ids

    @staticmethod
    def run_hb_method(
        data: pd.DataFrame,
        field_id: str,
        p_c: int,
        p_u: float,
        p_a: float,
        filter_op: str,
        filter_value: int,
    ) -> Figure:
        """Runs the HB-method on the data.

        :param data: Dataframe with the data
        :param field_id: Field ID
        :param p_c: pC value
        :param p_u: pU value
        :param p_a: pA value
        :param filter_op: Filter operator
        :param filter_value: Filter value
        :return: Figure for the scatterplot
        """
        hb_metoden_df = hb_method(
            data=data,
            id_field_name="id",
            x_1_field_name="x1",
            x_2_field_name="x2",
            p_c=p_c,
            p_u=p_u,
            p_a=p_a,
        )

        df_query = f"outlier == 1 & maxX {filter_op} {filter_value}"
        significant_outliers = hb_metoden_df.query(df_query)
        significant_outliers = significant_outliers.sort_values(by=["maxX"])

        x = significant_outliers["maxX"]
        z = significant_outliers["upperLimit"]
        k = significant_outliers["lowerLimit"]

        scatter_plot = px.scatter(
            significant_outliers,
            x="maxX",
            y="ratio",
            title=f"Post {field_id} - outliers med HB-metoden ({len(significant_outliers)} stk.)",
        )

        scatter_plot.add_trace(
            Scatter(
                x=x,
                y=z,
                name="Øvre grense",
                mode="lines",
                line=scatter.Line(color="yellow"),
                showlegend=True,
            )
        )

        scatter_plot.add_trace(
            Scatter(
                x=x,
                y=k,
                name="Nedre grense",
                mode="lines",
                line=scatter.Line(color="red"),
                showlegend=True,
            )
        )

        hover_template = (
            "Norsk-ID: <b>%{customdata}</b>"
            "<br>Beløp i hele 1000: <b>%{x}</b>"
            "<br>Forholdstall (ratio): <b>%{y}</b>"
        )

        scatter_plot.update_traces(
            customdata=significant_outliers["id"], hovertemplate=hover_template
        )

        scatter_plot.update_layout(
            xaxis_title="Beløp i hele 1000",
            yaxis_title="Forholdstall",
            plot_bgcolor="#1F2833",
            paper_bgcolor="#1F2833",
            font_color="#66FCF1",
            xaxis=dict(color="#66FCF1", hoverformat=",.2f"),
            yaxis=dict(color="#66FCF1", hoverformat=",.2f"),
        )
        return scatter_plot

    @staticmethod
    def empty_scatter_plot() -> Figure:
        """Creates a blank figure for use when no data is available.

        :return: Figure
        """
        empty_scatter_plot = px.scatter(x=[])
        empty_scatter_plot.update_xaxes(
            showgrid=False, showticklabels=False, zeroline=False
        )
        empty_scatter_plot.update_yaxes(
            showgrid=False, showticklabels=False, zeroline=False
        )
        empty_scatter_plot.update_layout(
            plot_bgcolor="#1F2833",
            paper_bgcolor="#1F2833",
            xaxis=dict(color="#1F2833"),
            yaxis=dict(color="#1F2833"),
        )

        return empty_scatter_plot

    def __init__(
        self,
        data: pd.DataFrame,
        field_id: str,
        id_field_name: str,
        x_1_name: str,
        x_2_name: str,
        default_p_c: int = 100,
        default_p_u: float = 0.5,
        default_p_a: float = 0.05,
        default_filter_op: str = "<",
        default_filter_value: int = 100_000,
        aio_id: Optional[str] = None,
    ) -> None:
        """Creates a new HbMethodAIO component.

        :param data:
        :param field_id:
        :param id_field_name:
        :param x_1_name:
        :param x_2_name:
        :param default_p_c:
        :param default_p_u:
        :param default_p_a:
        :param default_filter_op:
        :param default_filter_value:
        :param aio_id:
        """
        self.aio_id = aio_id if aio_id else str(uuid.uuid4())

        th_result = th_error(
            data=data,
            id_field_name=id_field_name,
            x_1_field_name=x_1_name,
            x_2_field_name=x_2_name,
        )

        data_dict = {"df": th_result.to_dict("records"), "field_id": field_id}

        super().__init__(
            children=[
                dcc.Store(data=data_dict, id=self.ids.data_store(self.aio_id)),
                dbc.Row(
                    id="hb-parameters",
                    className="m-2",
                    children=[
                        dbc.Col(
                            html.Span("HB-parametere (pC, pU, pA):"),
                            className="align-self-center",
                        ),
                        dbc.Col(
                            dbc.Select(
                                id=self.ids.p_c(self.aio_id),
                                options=[
                                    5,
                                    10,
                                    15,
                                    20,
                                    25,
                                    30,
                                    35,
                                    40,
                                    45,
                                    50,
                                    100,
                                    150,
                                ],
                                value=default_p_c,
                                className="m-2",
                                size="sm",
                                persistence=True,
                            )
                        ),
                        dbc.Col(
                            dbc.Select(
                                id=self.ids.p_u(self.aio_id),
                                options=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9],
                                value=default_p_u,
                                className="m-2",
                                size="sm",
                                persistence=True,
                            )
                        ),
                        dbc.Col(
                            dbc.Select(
                                id=self.ids.p_a(self.aio_id),
                                options=[
                                    0.01,
                                    0.02,
                                    0.03,
                                    0.04,
                                    0.05,
                                    0.06,
                                    0.07,
                                    0.08,
                                    0.09,
                                ],
                                value=default_p_a,
                                className="m-2",
                                size="sm",
                                persistence=True,
                            )
                        ),
                    ],
                ),
                dbc.Row(
                    id="filter-options",
                    className="m-2",
                    children=[
                        dbc.Col(
                            html.Span("Filter (etter HB):"),
                            className="align-self-center",
                        ),
                        dbc.Col(
                            dbc.Select(
                                id=self.ids.hb_filter_op(self.aio_id),
                                options=[
                                    {"label": "Større enn", "value": ">"},
                                    {"label": "Mindre enn", "value": "<"},
                                ],
                                value=default_filter_op,
                                className="m-2",
                                size="sm",
                                persistence=True,
                            )
                        ),
                        dbc.Col(
                            dbc.Select(
                                id=self.ids.hb_filter_value(self.aio_id),
                                options=[
                                    {"label": "100 MNOK", "value": "100000"},
                                    {"label": "10 MNOK", "value": "10000"},
                                    {"label": "1 MNOK", "value": "1000"},
                                    {"label": "0,1 MNOK", "value": "100"},
                                ],
                                value=default_filter_value,
                                className="m-2",
                                size="sm",
                                persistence=True,
                            )
                        ),
                    ],
                ),
                dcc.Graph(
                    id=self.ids.scatterplot(self.aio_id),
                    figure=HbMethodAIO.empty_scatter_plot(),
                ),
            ]
        )

    # noinspection PyMethodParameters
    @callback(
        Output(ids.scatterplot(MATCH), "figure"),
        [
            Input(ids.p_c(MATCH), "value"),
            Input(ids.p_u(MATCH), "value"),
            Input(ids.p_a(MATCH), "value"),
            Input(ids.hb_filter_op(MATCH), "value"),
            Input(ids.hb_filter_value(MATCH), "value"),
            State(ids.data_store(MATCH), "data"),
        ],
    )  # type: ignore
    def update_figures_cb(
        p_c: str,
        p_u: str,
        p_a: str,
        filter_operator: str,
        filter_value: str,
        data: dict[str, Any],
    ) -> Figure:
        """Updates the scatterplot figure.

        :param p_c: pC value
        :param p_u: pU value
        :param p_a: pA value

        :param filter_operator:
        :param filter_value:

        :param data: data from the store

        :return: figure for the scatterplot
        """
        return HbMethodAIO.run_hb_method(
            data=pd.DataFrame(data["df"]),
            field_id=data["field_id"],
            p_c=int(p_c),
            p_u=float(p_u),
            p_a=float(p_a),
            filter_op=str(filter_operator),
            filter_value=int(filter_value),
        )



import datetime
import json
from enum import Enum
from functools import wraps
from typing import Any
from typing import Optional

import dapla as dp
import pandas as pd

error_list = []
corrections_list = []
kontroll_dokumentasjon = {}


class Kontrolltype(Enum):
    """Enum class representing the types for quality control checks."""

    AUTOMATISK_OPPRETTING = 0
    MISSING = 1
    MULIG_FEIL = 2
    ABSOLUTT_FEIL = 3


class Feilrapport:
    """A class to detail errors found during quality control checks.

    Attributes:
        sub_control_id (str): Identifier for the sub control check.
        result_type (QualityControlResultType): The result type of the quality control check.
        context_id (str): Identifier for the context within which the error was detected.
        error_description (Optional[str]): A description of the error, if applicable.
    """

    def __init__(
        self,
        sub_control_id: str,
        result_type: Kontrolltype,
        context_id: str,
        error_description: Optional[str] = None,
        important_variables: Optional[list[str]] = None,
    ):
        self.sub_control_id = sub_control_id
        self.result_type = result_type
        self.context_id = context_id
        self.error_description = error_description
        self.important_variables = important_variables

    def to_dict(self) -> dict[str, Any]:
        """Converts the quality control result into a dictionary format.

        Returns:
            A dictionary representing the quality control result.
        """
        return {
            "kontrollnavn": self.sub_control_id,
            "kontrolltype": self.result_type.name,
            "observasjon_id": self.context_id,
            "feilbeskrivelse": self.error_description,
            "relevante_variabler": self.important_variables,
        }


def kontroll(
    result_type: Kontrolltype,
    error_description: str,
    id_column: str,
    filter_for_relevant_data=None,
    important_variables: list | None = None,
):
    """Assumes all data given to the function is data that should be checked

    Note:
    """

    def decorator(control_function):
        @wraps(control_function)
        def wrapper(data: pd.DataFrame):
            data = data.copy()
            if "utslag" in data.columns:
                data = data.drop(columns=["utslag"])
                print("Droppet eksisterende utslag")
            if filter_for_relevant_data:
                total_data = data.shape[0]
                data = filter_for_relevant_data(data)
                if data.shape[0] == total_data:
                    print("Warning! Data before and after filter are the same length")
            else:
                print(
                    "Warning! Report will show 'Enheter i datasettet' and 'Enheter kontorllert' as the same amount"
                )
                total_data = None

            data["utslag"] = False
            filtered_data = control_function(data)
            error_rows = filtered_data[filtered_data["utslag"]]

            new_error_details = [
                Feilrapport(
                    sub_control_id=control_function.__name__,
                    result_type=result_type,
                    context_id=row[id_column],
                    error_description=error_description,
                    important_variables=important_variables,
                )
                for index, row in error_rows.iterrows()
            ]

            global error_list
            error_list.extend(new_error_details)

            global kontroll_dokumentasjon
            kontroll_dokumentasjon[control_function.__name__] = {
                "kontrolltype": result_type.name,
                "feilbeskrivelse": error_description,
                "docstring": control_function.__doc__,
                # Må finne en annen måte å definere antall enheter i datasett og antall enheter kontrollert
                "Enheter i datasettet": total_data if total_data else data.shape[0],
                "Enheter kontrollert": data.shape[0],
                "Kontrollutslag": error_rows.shape[0],
            }
            if important_variables:
                kontroll_dokumentasjon[control_function.__name__][
                    "Relevante variabler"
                ] = important_variables

            data = data.drop(columns=["utslag"])

            return data

        return wrapper

    return decorator


def automatisk_oppretting(id_column: str, correction_description: str):
    def decorator(correction_function):
        @wraps(correction_function)
        def wrapper(data: pd.DataFrame):
            data = data.copy()
            corrected_data = correction_function(data)

            corrected_rows = corrected_data[corrected_data["maskinelt_rettet"]]

            new_error_details = [
                Feilrapport(
                    sub_control_id=correction_function.__name__,
                    result_type=Kontrolltype.AUTOMATISK_OPPRETTING,
                    context_id=row[id_column],
                    error_description=correction_description,
                )
                for index, row in corrected_rows.iterrows()
            ]

            global corrections_list
            corrections_list.extend(new_error_details)

            global kontroll_dokumentasjon
            kontroll_dokumentasjon[correction_function.__name__] = {
                "kontrolltype": Kontrolltype.AUTOMATISK_OPPRETTING,
                "feilbeskrivelse": correction_description,
                "docstring": correction_function.__doc__,
                # Må finne en annen måte å definere antall enheter i datasett og antall enheter kontrollert
                "Enheter i datasettet": data.shape[0],
                "Enheter kontrollert": data.shape[0],
                "Kontrollutslag": corrected_rows.shape[0],
            }

            return corrected_data

        return wrapper

    return decorator


class Kvalitetsrapport:
    """A class representing the result of a quality control check.

    Attributes:
        statistics_name (str): The name of the statistics being checked.
        quality_control_id (str): The unique identifier for the quality control.
        data_location (list[str]): Locations of the data checked.
        data_period (str): The period for which the data was checked.
        quality_control_datetime (datetime): The datetime when the quality control was performed.
        quality_control_results (list[QualityControlResultType]): The results of the quality control.
        quality_control_errors (list[QualityReportErrorDetails]): Detailed errors found during the quality control.
    """

    def __init__(
        self,
        statistics_name: str,
        quality_control_id: str,
        data_location: list[str],
        data_period: str,
        quality_control_datetime: datetime,
        quality_control_results: list[Kontrolltype],
        quality_control_errors: list[Feilrapport],
        corrections=list[Feilrapport],
    ):
        self.statistics_name = statistics_name
        self.quality_control_id = quality_control_id
        self.data_location = data_location
        self.data_period = data_period
        self.quality_control_datetime = quality_control_datetime
        self.quality_control_results = quality_control_results
        self.quality_control_errors = quality_control_errors
        self.corrections = corrections

    def to_dict(self) -> dict[str, Any]:
        """Converts the quality control result into a dictionary format.

        Returns:
            A dictionary representing the quality control result.
        """
        return {
            "statistikknavn": self.statistics_name,
            "quality_control_id": self.quality_control_id,  # Trengs denne?
            "data_plassering": self.data_location,
            "data_periode": self.data_period,
            "kvalitetsrapport opprettet": self.quality_control_datetime.isoformat(),
            "typer_kontrollutslag": [
                result.name for result in self.quality_control_results
            ],
            "kontrollutslag": [
                error.to_dict() for error in self.quality_control_errors
            ],
            "korreksjoner": [error.to_dict() for error in self.quality_control_errors],
            "kontrolldokumentasjon": kontroll_dokumentasjon,
        }

    def save_report(self, path):
        with dp.FileClient.gcs_open(path, "w") as outfile:
            json.dump(self.to_dict(), outfile)


def lag_kvalitetsrapport(
    statistics_name,
    data_location,
    data_period,
    also_return_control_docs=False,
):
    """ """
    assert statistics_name is not None, "Må sette statistikknavn (statistics_name)"
    assert data_location is not None, "Mangler filsti til datasett (data_location)"
    assert (
        data_period is not None
    ), "Må definere hvilken periode dataene gjelder for (data_period)"

    control_errors_nested = [error_list]

    control_errors = [item for sublist in control_errors_nested for item in sublist]

    quality_results: list[Kontrolltype] = []

    for i in Kontrolltype:
        if any(error.result_type == i for error in control_errors):
            quality_results.append(i)
    if not any(quality_results):
        quality_results.append(Kontrolltype.OK)

    report = Kvalitetsrapport(
        statistics_name=statistics_name,
        quality_control_id="A reference (or link/uri) to the quality control description",
        data_location=[data_location],
        data_period=data_period,
        quality_control_datetime=datetime.datetime.now(),
        quality_control_results=quality_results,
        quality_control_errors=error_list,
        corrections=corrections_list,
    )

    if also_return_control_docs:
        return report, lag_kontroll_dokumentasjon(report)
    else:
        return report


def lag_kontroll_dokumentasjon(kvalitetsrapport):
    if isinstance(kvalitetsrapport, dict):
        kontrolldokumentasjon = pd.DataFrame(
            kvalitetsrapport["kontrolldokumentasjon"]
        ).T.assign(periode=kvalitetsrapport["data_periode"])
    elif isinstance(kvalitetsrapport, Kvalitetsrapport):
        kontrolldokumentasjon = pd.DataFrame(
            kvalitetsrapport.to_dict()["kontrolldokumentasjon"]
        ).T.assign(periode=kvalitetsrapport.to_dict()["data_periode"])
    kontrolldokumentasjon.index.names = ["kontroll_id"]
    kontrolldokumentasjon = kontrolldokumentasjon.reset_index()
    return kontrolldokumentasjon


def eimerdb_template(kontrolldokumentasjon):
    kontroller = []
    for i in kontrolldokumentasjon.to_dict()["kontrolldokumentasjon"]:
        kontroller.append(
            [
                kontrolldokumentasjon.to_dict()["data_periode"],
                i,
                kontrolldokumentasjon.to_dict()["kontrolldokumentasjon"][i][
                    "kontrolltype"
                ].name,
                kontrolldokumentasjon.to_dict()["kontrolldokumentasjon"][i][
                    "feilbeskrivelse"
                ],
            ]
        )
    return kontroller



import dash_bootstrap_components as dbc
from dash import html


def sidebar_button(icon, text, component_id, additional_styling=None):
    if additional_styling is None:
        additional_styling = {}
    button = html.Div(
        dbc.Button(
            [
                html.Span(icon, style={"display": "block", "font-size": "1.4rem"}),
                html.Span(text, style={"display": "block", "font-size": "0.7rem"}),
            ],
            id=component_id,
            style={
                "display": "flex",
                "flex-direction": "column",
                "align-items": "center",
                "word-break": "break-all",
                "margin-bottom": "5%",
                "width": "100%",
                **additional_styling,
            },
        )
    )
    return button



import dash_bootstrap_components as dbc
from dash import Dash
from dash_bootstrap_templates import load_figure_template

theme_map = {
    "sketchy": dbc.themes.SKETCHY,
    "slate": dbc.themes.SLATE,
    "cyborg": dbc.themes.CYBORG,
    "superhero": dbc.themes.SUPERHERO,
    "darkly": dbc.themes.DARKLY,
    "solar": dbc.themes.SOLAR,
    "flatly": dbc.themes.FLATLY,
}


def app_setup(port, service_prefix, domain, stylesheet):
    template = theme_map[stylesheet]
    load_figure_template([template])

    dbc_css = (
        "https://cdn.jsdelivr.net/gh/AnnMarieW/dash-bootstrap-templates/dbc.min.css"
    )

    app = Dash(
        __name__,
        requests_pathname_prefix=f"{service_prefix}proxy/{port}/",
        external_stylesheets=[theme_map[stylesheet], dbc_css],
    )
    return app



import dash_ag_grid as dag
from dash import callback
from dash import dcc
from dash import html
from dash.dependencies import Input
from dash.dependencies import Output
from dash.dependencies import State

input_options = {
    "orgb": Input("var-bedrift", "value"),
    "orgf": Input("var-foretak", "value"),
    "oppgavegiver": Input("var-oppgavegiver", "value"),
    "skjemaident": Input("var-skjemaident", "value"),
}

states_options = [
    {
        "aar": ("var-aar", "value"),
        "termin": ("var-termin", "value"),
        "nace": ("var-nace", "value"),
    }
]


class EditingTable:
    def __init__(
        self,
        label,
        database,
        tables,
        var_input,
        states,
        get_data_func,
        update_table_func,
    ):
        dropdown_options = [{"label": table, "value": table} for table in tables]

        self.selected_ident = input_options[var_input]
        self.states = states
        self.get_data = get_data_func
        self.update_table = update_table_func
        self.database = database
        self.dropdown_options = dropdown_options
        self.callbacks()
        self.label = label

    def layout(self):
        layout = html.Div(
            style={"height": "100vh", "display": "flex", "flexDirection": "column"},
            children=[
                html.Div(
                    children=[
                        dcc.Dropdown(
                            id="tab-tabelleditering-dd1",
                            options=self.dropdown_options,
                            value=self.dropdown_options[0]["value"],
                            placeholder="Velg tabell",
                            className="dbc",
                        ),
                        dag.AgGrid(
                            defaultColDef={"editable": True},
                            id="tab-tabelleditering-table1",
                            className="ag-theme-alpine-dark header-style-on-filter",
                        ),
                        html.P(id="tab-tabelleditering-status1"),
                    ],
                ),
            ],
        )
        return layout

    def callbacks(self):
        states_dict = states_options[0]
        dynamic_states = [
            State(states_dict[key][0], states_dict[key][1]) for key in self.states
        ]

        @callback(
            Output("tab-tabelleditering-table1", "rowData"),
            Output("tab-tabelleditering-table1", "columnDefs"),
            Input("tab-tabelleditering-dd1", "value"),
            self.selected_ident,
            *dynamic_states,
        )
        def load_ag_grid(tabell, ident, *dynamic_states):
            states_values = dynamic_states[: len(self.states)]
            state_params = {
                key: value for key, value in zip(self.states, states_values)
            }

            args = []
            for key in self.states:
                var = state_params.get(key)
                if var is not None:
                    args.append(var)
            if "nace" in state_params and state_params["nace"] is not None:
                n = len(state_params["nace"])
                args.append(n)
            df = self.get_data(self.database, ident, tabell, *args)
            columns = [
                {
                    "headerName": col,
                    "field": col,
                    "hide": True if col == "row_id" else False,
                }
                for col in df.columns
            ]
            columns[0]["checkboxSelection"] = True
            columns[0]["headerCheckboxSelection"] = True
            return df.to_dict("records"), columns

        @callback(
            Output("tab-tabelleditering-status1", "children"),
            Input("tab-tabelleditering-table1", "cellValueChanged"),
            State("tab-tabelleditering-dd1", "value"),
            *dynamic_states,
        )
        def update_table(edited, tabell, *dynamic_states):
            states_values = dynamic_states[: len(self.states)]
            state_params = {
                key: value for key, value in zip(self.states, states_values)
            }

            args = []
            for key in self.states:
                var = state_params.get(key)
                if var is not None:
                    args.append(var)
            variable = edited[0]["colId"]
            old_value = edited[0]["oldValue"]
            new_value = edited[0]["value"]
            row_id = edited[0]["data"]["row_id"]
            try:
                self.update_table(
                    self.database, variable, new_value, row_id, tabell, *args
                )
                return f"{variable} updatert fra {old_value} til {new_value}"
            except:
                return "Error"



"""SSB Sirius Dash."""

import importlib
import os

# Path to the src folder
src_dir = os.path.dirname(__file__)

# Traverse the modules in the ssb-sirius-dash directory
for root, dirs, files in os.walk(src_dir):
    for file in files:
        if file.endswith(".py") and not file.startswith("_"):
            module_name = os.path.splitext(file)[0]
            module_path = os.path.relpath(os.path.join(root, file), src_dir)
            module_path = module_path.replace(os.sep, ".").replace(".py", "")

            # Import the module dynamically
            module = importlib.import_module(
                f".{module_path}", package="ssb_sirius_dash"
            )

            # Get all attributes that don't start with "_"
            for attr_name in dir(module):
                if not attr_name.startswith("_"):
                    globals()[attr_name] = getattr(module, attr_name)



import uuid
from typing import Any
from typing import Optional

import pandas as pd
import plotly.express as px
from dash.dcc import Graph


class SunburstAIO(Graph):  # type: ignore
    """SunburstAIO is an All-in-One component that is composed."""

    class Ids:
        """A set of functions that create pattern-matching callbacks of the subcomponents."""

        @staticmethod
        def sunburst(aio_id: Any) -> dict[str, Any]:
            """Returns the component ID for the sunburst component.

            :param aio_id: The All-in-One component ID
            :return: The component ID for the sunburst component
            """
            return {
                "component": "SunburstAIO",
                "subcomponent": "sunburst",
                "aio_id": aio_id,
            }

    ids = Ids

    def __init__(
        self,
        data: pd.DataFrame,
        path: list[str],
        values: str,
        aio_id: Optional[str] = None,
    ) -> None:
        """SunburstAIO is an All-in-One component that is composed.

        :param data: DataFrame
        :param path: Path to show in the sunburst
        :param values: Name of value column
        :param aio_id: The All-in-One component ID used to generate the sunburst component's dictionary IDs.
        """
        self.aio_id = aio_id if aio_id else str(uuid.uuid4())

        figure = px.sunburst(data, path=path, values=values)

        figure.update_layout(
            plot_bgcolor="#1F2833",
            paper_bgcolor="#1F2833",
            font_color="#66FCF1",
            xaxis=dict(color="#66FCF1"),
            yaxis=dict(color="#66FCF1"),
        )

        super().__init__(id=self.ids.sunburst(self.aio_id), figure=figure)



import dash_bootstrap_components as dbc
from dash import html

variable_options = {
    "aar": {"title": "År", "id": "var-aar", "type": "number"},
    "termin": {"title": "Termin", "id": "var-termin", "type": "number"},
    "nace": {"title": "Nace", "id": "var-nace", "type": "text"},
    "oppgavegiver": {"title": "Oppgavegiver", "id": "var-oppgavegiver", "type": "text"},
    "foretak": {"title": "Foretak", "id": "var-foretak", "type": "text"},
    "bedrift": {"title": "Bedrift", "id": "var-bedrift", "type": "text"},
    "fylke": {"title": "Fylke", "id": "var-fylke", "type": "text"},
    "skjemaenhet": {"title": "Skjemaenhet", "id": "var-skjemaenhet", "type": "text"},
    "prodcomkode": {"title": "Prodcomkode", "id": "var-prodcomkode", "type": "text"},
    "nspekfelt": {"title": "NSPEK-felt", "id": "var-nspekfelt", "type": "text"},
}


def skjermcard(text, component_id, input_type, value=None):
    if value is None:
        value = ""
    card = dbc.Col(
        dbc.Card(
            dbc.CardBody(
                [
                    html.H5(text, className="card-title"),
                    html.Div(
                        style={
                            "display": "grid",
                            "grid-template-columns": "100%",
                        },
                        children=[
                            dbc.Input(value=value, id=component_id, type=input_type),
                        ],
                    ),
                ],
                style={"max-height": "100%"},
            ),
            style={"max-height": "100%"},
        )
    )
    return card


def generate_skjermcards(selected_keys, default_values=None):
    if default_values is None:
        default_values = {}
    cards_list = []
    for key in selected_keys:
        card_config = variable_options.get(key, {})
        title = card_config.get("title")
        card_id = card_config.get("id")
        card_type = card_config.get("type")
        value = default_values.get(key, None)

        card = skjermcard(title, card_id, card_type, value)
        cards_list.append(card)
    return cards_list
